from typed_args import TypedArgs, add_argument
from dataclasses import dataclass
from typing import Optional


def test_optional_argument():
    @dataclass
    class Args(TypedArgs):
        config: Optional[str] = add_argument('-c', '--config')

    args = Args.from_args()

    assert args.config == None

    args = Args.from_args('-c /path'.split())

    assert args.config == '/path'


def test_optional_with_default():
    @dataclass
    class Args(TypedArgs):
        config: str = add_argument('-c', '--config', default='/path')

    args = Args.from_args()

    assert args.config == '/path'

    args = Args.from_args('-c /other'.split())

    assert args.config == '/other'



from dataclasses import dataclass

from typed_args import TypedArgs, add_argument


@dataclass
class Args(TypedArgs):
    data: str = add_argument(metavar='DIR', help='path to dataset')
    arch: str = add_argument('-a', '--arch', metavar='ARCH', default='resnet18',
                             help='model architecture (default: resnet18)')
    num_workers: int = add_argument('-j', '--workers', default=4, metavar='N',
                                    help='number of data loading workers (default: 4)')


def test_args():
    data = '/path/to/dataset'
    arch = 'resnet50'
    num_workers = 8

    argv = f'{data} -a {arch} --workers {num_workers}'.split()

    # sys.argv.extend(argv)

    args = Args.from_known_args(argv)

    assert args.arch == arch
    assert args.data == data
    assert args.num_workers == num_workers


def test_known_args():
    data = '/path/to/dataset'
    arch = 'resnet50'
    num_workers = 8

    argv = f'{data} -a {arch} --workers {num_workers}'.split()

    # sys.argv.extend(argv)

    args = Args.from_args(argv)

    assert args.arch == arch
    assert args.data == data
    assert args.num_workers == num_workers


if __name__ == "__main__":
    # test_args()
    test_known_args()



from typed_args import TypedArgs, add_argument, dataclass
from typing import List


def test_list():
    """
    https://docs.python.org/3/library/argparse.html#nargs
    :return:
    """

    @dataclass
    class Args(TypedArgs):
        foo: List[str] = add_argument('--foo', nargs=2)
        bar: List[str] = add_argument(nargs=1)

    args = Args.from_args('c --foo a b'.split())

    assert args.foo == ['a', 'b']
    assert args.bar == ['c']



import typing
import collections


def get_origin(tp):
    """Get the unsubscripted version of a type.
    This supports generic types, Callable, Tuple, Union, Literal, Final and ClassVar.
    Return None for unsupported types. Examples::
        get_origin(Literal[42]) is Literal
        get_origin(int) is None
        get_origin(ClassVar[int]) is ClassVar
        get_origin(Generic) is Generic
        get_origin(Generic[T]) is Generic
        get_origin(Union[T, int]) is Union
        get_origin(List[Tuple[T, T]][int]) == list
    """
    if isinstance(tp, typing._GenericAlias):
        return tp.__origin__
    if tp is typing.Generic:
        return typing.Generic
    return None


def get_args(tp):
    """Get type arguments with all substitutions performed.
    For unions, basic simplifications used by Union constructor are performed.
    Examples::
        get_args(Dict[str, int]) == (str, int)
        get_args(int) == ()
        get_args(Union[int, Union[T, int], str][int]) == (int, str)
        get_args(Union[int, Tuple[T, int]][str]) == (int, Tuple[str, int])
        get_args(Callable[[], T][int]) == ([], int)
    """
    if isinstance(tp, typing._GenericAlias):
        res = tp.__args__
        if get_origin(tp) is collections.abc.Callable and res[0] is not Ellipsis:
            res = (list(res[:-1]), res[-1])
        return res
    return ()



import logging
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass, field
from typing import Union, Optional, Any, Iterable, List, Tuple

try:
    # Python 3.8
    from typing import get_origin, get_args
except ImportError:
    from .utils import get_origin, get_args

__version__ = '0.3.3'

LOGGER = logging.getLogger(__name__)


@dataclass
class TypedArgs:
    parser: ArgumentParser = field(default_factory=ArgumentParser)

    @classmethod
    def from_args(cls, args: Optional[List[str]] = None, namespace: Optional[Namespace] = None):
        typed_args = cls()
        typed_args.parse_args(args=args, namespace=namespace)
        return typed_args

    @classmethod
    def from_known_args(cls, args: Optional[List[str]] = None, namespace: Optional[Namespace] = None):
        typed_args = cls()
        typed_args.parse_known_args(args=args, namespace=namespace)
        return typed_args

    def add_arguments(self):
        for name, annotation in self.__annotations__.items():
            if name == 'parser':
                continue
            self.add_argument(name, annotation)

    def add_argument(self, name: str, annotation: Any):
        values: Union[PhantomAction, Tuple[PhantomAction]] = getattr(self, name)

        if type(values) != tuple:
            types = (annotation,)
            values = (values,)
        else:
            types = annotation.__args__[0].__args__
            if types[-1] == Ellipsis:
                types = (types[0],) * len(values)

        for argument_type, value in zip(types, values):
            kwargs = value.to_kwargs()
            args = kwargs.pop('option_strings', ())

            kwargs['dest'] = name  # 必定有dest

            is_position_argument = len(args) == 0

            origin = get_origin(argument_type)

            if origin is list:
                argument_type = get_args(argument_type)[0]

                print('argument type: ', argument_type)

            if kwargs['action'] == 'store':

                # 不存在default的才需要判断optional
                if kwargs.get('default', None) is None:

                    if origin is Union:  # Optional
                        argument_type = get_args(argument_type)[0]
                        kwargs['required'] = False
                    elif origin is None:
                        if not is_position_argument:
                            kwargs['required'] = True

                kwargs['type'] = argument_type

            self.parser.add_argument(*args, **kwargs)

    def parse_args(self, args: Optional[List[str]] = None, namespace: Optional[Namespace] = None):
        self.add_arguments()
        parsed_args = self.parser.parse_args(args=args, namespace=namespace)
        self.update_arguments(parsed_args)

    def parse_known_args(self, args: Optional[List[str]] = None, namespace: Optional[Namespace] = None):
        self.add_arguments()
        parsed_args, _ = self.parser.parse_known_args(
            args=args, namespace=namespace)
        self.update_arguments(parsed_args)

    def update_arguments(self, parsed_args: Namespace):
        for name in self.__annotations__.keys():
            if name == 'parser':
                continue
            value = getattr(parsed_args, name)
            setattr(self, name, value)


@dataclass
class PhantomAction:
    option_strings: Tuple[str, ...]
    action: str = 'store'
    nargs: Union[int, str, None] = None
    const: Any = None
    default: Any = None
    choices: Optional[Iterable] = None
    required: Optional[bool] = None
    help: Optional[str] = None
    metavar: Optional[str] = None

    def to_kwargs(self):
        kwargs = self.__dict__.copy()

        if len(self.option_strings) == 0:
            # position argument is always required
            del kwargs['required']

        if self.action.startswith('store_'):
            del kwargs['nargs']
            del kwargs['choices']

            if self.action in ['store_true', 'store_false']:
                del kwargs['metavar']
                del kwargs['const']
                del kwargs['default']

        elif self.action == 'append_const':
            del kwargs['nargs']
            del kwargs['choices']

        elif self.action == 'count':
            del kwargs['nargs']
            del kwargs['choices']
            del kwargs['const']
            del kwargs['metavar']

        elif self.action == 'help':
            del kwargs['dest']
            del kwargs['default']

            del kwargs['nargs']
            del kwargs['choices']
            del kwargs['const']
            del kwargs['metavar']
            del kwargs['required']

        elif self.action == 'version':
            del kwargs['dest']
            del kwargs['default']

            del kwargs['nargs']
            del kwargs['choices']
            del kwargs['const']
            del kwargs['metavar']
            del kwargs['required']

        elif self.action == 'parsers':
            raise NotImplemented()

        return kwargs


def add_argument(
        *option_strings: Union[str, Tuple[str, ...]],
        action: str = 'store',  # in argparse, default action is 'store'
        nargs: Union[int, str, None] = None,
        const: Optional[Any] = None,
        default: Optional[Any] = None,
        choices: Optional[Iterable] = None,
        required: Optional[bool] = None,
        help: Optional[str] = None,
        metavar: Optional[str] = None,
):
    """

    :param option_strings:
    :param action:
    :param nargs:
    :param const:
    :param default:
    :param choices:
    :param required:
    :param help:
    :param metavar:
    :return:
    """
    kwargs = locals()
    LOGGER.debug('local = ', kwargs)

    # print('=' * 100)
    # print(kwargs)
    # print('=' * 100)

    return PhantomAction(**kwargs)



import io
import os
import re

from setuptools import find_packages, setup


def read(*names, **kwargs):
    with io.open(
            os.path.join(os.path.dirname(__file__), *names),
            encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    else:
        raise RuntimeError("Unable to find version string.")


def read_long_description():
    with open('README.md', 'r') as f:
        long_description = f.read()

    return long_description


VERSION = find_version('typed_args', '__init__.py')

requirements = [
]

setup(
    name='typed-args',
    version=VERSION,
    author='SunDoge',
    author_email='384813529@qq.com',
    long_description=read_long_description(),
    long_description_content_type='text/markdown',
    url='https://github.com/SunDoge/typedargs',
    packages=find_packages(exclude=['test']),
    install_requires=requirements,
)



from dataclasses import dataclass
from typing import List, Union

from typed_args import TypedArgs, add_argument


def test_append():
    @dataclass
    class Args(TypedArgs):
        foo: List[str] = add_argument('--foo', action='append')

    args = Args.from_args('--foo 1 --foo 2'.split())

    assert args.foo == ['1', '2']


def test_same_dest():
    @dataclass
    class Args(TypedArgs):
        types: List[Union[str, int]] = (
            add_argument('--str', action='append_const', const=str),
            add_argument('--int', action='append_const', const=int)
        )

    # parser = argparse.ArgumentParser()
    # parser.add_argument('--str', dest='types', action='append_const', const=str)
    # parser.add_argument('--int', dest='types', action='append_const', const=int)

    args = Args.from_args('--str --int'.split())
    # args = parser.parse_args('--str --int'.split())
    # args = parser.parse_args('--int --str'.split())

    assert args.types == [str, int]



from typed_args import TypedArgs, add_argument
from dataclasses import dataclass, field
from typing import *
import argparse


def test_name_or_flags():
    """
    https://docs.python.org/3/library/argparse.html#name-or-flags

    :return:
    """
    import argparse
    @dataclass
    class Args(TypedArgs):
        parser: argparse.ArgumentParser = field(default_factory=lambda: argparse.ArgumentParser(prog='PROG'))
        foo: Optional[str] = add_argument('-f', '--foo')
        bar: str = add_argument()

    args = Args.from_args(['BAR'])

    assert args.bar == 'BAR'
    assert args.foo == None

    args = Args.from_args(['BAR', '--foo', 'FOO'])

    assert args.bar == 'BAR'
    assert args.foo == 'FOO'


def test_action_store():
    @dataclass
    class Args1(TypedArgs):
        foo: Optional[str] = add_argument('--foo')

    args1 = Args1.from_args('--foo 1'.split())
    assert args1.foo == '1'


def test_action_score_const():
    @dataclass
    class Args2(TypedArgs):
        foo: int = add_argument('--foo', action='store_const', const=42)

    args2 = Args2.from_args(['--foo'])
    assert args2.foo == 42


def test_action_store_true_false():
    @dataclass
    class Args3(TypedArgs):
        foo: bool = add_argument('--foo', action='store_true')
        bar: bool = add_argument('--bar', action='store_false')
        baz: bool = add_argument('--baz', action='store_false')

    args3 = Args3.from_args('--foo --bar'.split())
    assert args3.foo == True
    assert args3.bar == False
    assert args3.baz == True


def test_action_append():
    @dataclass
    class Args4(TypedArgs):
        foo: List[str] = add_argument('--foo', action='append')

    args4 = Args4.from_args('--foo 1 --foo 2'.split())
    assert args4.foo == ['1', '2']


def test_action_append_const():
    @dataclass
    class Args5(TypedArgs):
        types: List[Union[str, int]] = (
            add_argument('--str', action='append_const', const=str),
            add_argument('--int', action='append_const', const=int)
        )

    args5 = Args5.from_args('--str --int'.split())
    assert args5.types == [str, int]


def test_action_count():
    @dataclass
    class Args(TypedArgs):
        verbose: int = add_argument('--verbose', '-v', action='count', default=0)

    args = Args.from_args(['-vvv'])
    assert args.verbose == 3


# TODO help, version, extend

def test_nargs_n():
    @dataclass
    class Args(TypedArgs):
        foo: List[str] = add_argument('--foo', nargs=2)
        bar: List[str] = add_argument(nargs=1)

    args = Args.from_args('c --foo a b'.split())
    assert args.bar == ['c']
    assert args.foo == ['a', 'b']


def test_nargs_optional():
    @dataclass
    class Args(TypedArgs):
        foo: str = add_argument('--foo', nargs='?', const='c', default='d')
        bar: str = add_argument(nargs='?', default='d')

    args = Args.from_args(['XX', '--foo', 'YY'])
    assert args.foo == 'YY'
    assert args.bar == 'XX'

    args = Args.from_args(['XX', '--foo'])
    assert args.foo == 'c'
    assert args.bar == 'XX'

    args = Args.from_args()
    assert args.foo == 'd'
    assert args.bar == 'd'
